'*********************************************************
'*                                                       *
'*   ZiCog_cpm v1.0_rc_6                                 *
'*                                                       *
'*   A simulation of a Zilog Z80 based computer system   *
'*   Capable of running CP/M (v2.2 and 3)                *
'*                                                       *
'*                                                       *
'*   (C) 2010 Michael Rychlik                            *
'*                                                       *
'*********************************************************
'
CON

  baud          = 115_200                               'Set your serial speed


'USER CONFIGURATION OPTIONS

'Uncomment one of the following CPU emulation types.
#define CPU_Z80
'#define CPU_8080

'Uncomment one of the following platform types.
#define TriBladeProp
'#define UseTriBladeSD                                   'Uncomment to run PropDemoBoard on TriBlade hardware
#define Boot_from_SD
'#define Boot_from_SD_SPIN


'#define banked_memory                                  'Uncomment if banked memory is emulated.(48K+16K)+4x48KB=256KB
'#define SingleStep                                     'Uncomment if "single step" emulation required.
'#define HEARTBEAT                                      'Uncomment if Z80 emulation is to flash a heartbeat LED
#define Show_Hex                                        'Uncomment to display in hex, else display char
'#define ShowDiskWrites                                  'Uncomment to display "." for each hdisk write


_clkmode      = xtal1 + pll16x
_xinfreq      = 6_553_600

cpmsize       = 64   * 1024                           'CPM Ram size (see msize below)
sramsize      = 512  * 1024                           'Uncomment for 512KB SRAM
'  sramsize      = 1024 * 1024                           'Uncomment for 1MB  SRAM


' microSD pins on TriBladeProp Blade #2
spiDO         = 9
spiClk        = 28
spiDI         = 8
spiCS         = 14

OBJ

  cpu      : "zicog"                                    'Zilog Z80 emulator in one COG
  UART     : "FullDuplexSerial_rr004"                   'FDX v1.2 with mods
#ifdef TriBladeProp
  tbp2     : "TriBladeProp_Blade2_202"                  'Blade2 hardware driver
#ifdef fsrw_new
  sd       : "fsrw_rr001"                               'SD access (latest from lonesock/rokicki)
#else
  sd       : "fsrwFemto_rr001"                          'SD Software used in FemtoBASIC
#endif
#else
  sd       : "sdspiqasm_rr001"                          'Low level access to SD card (passes all pins)
#endif


CON

'Z80 Program Status Word bits
sign_bit        = %10000000
zero_bit        = %01000000
aux_bit         = %00010000
parity_bit      = %00000100
carry_bit       = %00000001


'I/O control block commands
io_cmd_out      = $01
io_cmd_in       = $02
io_cmd_break    = $03

#ifdef FloppySupport
' commands for disk controller (dskcon)
cstepin         = $01                   ' step in command
cstepot         = $02                   ' step out command
cload           = $04                   ' load head to drive surface command
cuload          = $08                   ' unload head from drive surface command
cwrseq          = $80                   ' 'start write enable sequence' command

ndisks          = 8                     ' total number of Altair disks
tracks          = 254                   ' number of tracks for regular drives
track1          = tracks+1              ' indicator for unknown track position
asecsiz         = 137                   ' sector size Altair
csecsiz         = $0080                 ' sector size CP/M

selout          = 8                     ' port to select and enable controller and drive (OUT)
statin          = 8                     ' port indicating status of drive and controller (IN)
dskcon          = 9                     ' port to control disc function (OUT)
secpos          = 9                     ' port to indicate current sector position of disk (IN)
dskwrit         = 10                    ' port to write data (OUT)
dskread         = 10                    ' port to read data (IN)
#endif



'
'       A read to the data port gets the buffered character, a write
'       to the data port writes the character to the device.
constat         = 16                    ' sio port 1 status port
condata         = 17                    ' sio port 1 data port
punstat         = 18                    ' sio port 2 status port
pundata         = 19                    ' sio port 2 data port

simhport        = $FE
srport          = $FF                  ' Altair front panel switch register port

ledport         = $00

' UART constants for N8VEM
n8vem_uart_test = $6D
n8vem_uart_in   = $68
n8vem_uart_out  = $68

' masks for SIO controller (constat, punstat)
mout            = $02                   ' output allowed mask

' commands for SIO controller (constat, punstat)
creset          = 3                     ' reset command

' constants for hard disk port
nhdisks         = 8                     'The number of supported hard disks

hdskNull        = 0                     ' No command
hdskReset       = 1                     ' command to reset controller
hdskRead        = 2                     ' read command
hdskWrite       = 3                     ' write command
hdskParam       = 4                     ' Get HD parameters command
hdskport        = $fd                   ' control port for simulated hard disk

#ifdef UseTriBladeSD
hdsk_base_block = $202c7                ' First block of first hard disk in SD "zipm2.dsk"
#else
hdsk_base_block = 0                     ' First block of first hard disk in SD.
#endif

'constants for SIMH port
simhreset       = 14
stoptimer       = 22
setbankselect   = 12
hasbankedmemory = 18

DAT

'The Z80 Registers.
'Do not change the order of these.
'Registers are defined in pairs as words so as to enable them to be
'accessed individually as BYTE or in pairs as WORD.
'The ordering makes it easier to find a register specified by a 3-bit
'register field within an op code or to swap between main and
'alternate register sets. 
register_file word
bc_reg        word      0
de_reg        word      0
hl_reg        word      0
af_reg        word      0
bc_reg_alt    word      0 
de_reg_alt    word      0
hl_reg_alt    word      0
af_reg_alt    word      0
im_reg        word      0
ix_reg        word      0
iy_reg        word      0
sp_reg        word      0
pc_reg        word      0
  
'IO interface area. Do not change the order of these.
io_cmd_block  long
io_command    long      0
io_port       long      0            
io_data       long      0


'Hard Disk Parameter Blocks
'Each DPB is 17 bytes as defined for DPBs in CP/M followed by 2 bytes sector size as used by SIMH
hdsk_dpb
dpb_0                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)
                                'N.B. SS must be 128 for CP/M 2 can be varied for CP/M 3 hard disks.

dpb_1                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determdirine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)
                                'N.B. SS must be 128 for CP/M 2 can be varied for CP/M 3 hard disks.

dpb_2                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)

dpb_3                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)

dpb_4                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)

dpb_5                           '8MByte Default SIMH Altair HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $05     'data allocation Block SHift factor
:blm          BYTE      $1F     'data allocation block mask
:exm          BYTE      $01     'EXtent Mask
:dsm_low      BYTE      $F9     'maximum data block number (low_byte)
:dsm_high     BYTE      $07     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $03     'total number of directory entries (high byte)
:al0          BYTE      $FF     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)

dpb_6                           '1MByte Altair mits2 floppy as HDSK params (no skew)
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $04     'data allocation Block SHift factor
:blm          BYTE      $0F     'data allocation block mask
:exm          BYTE      $00     'EXtent Mask
:dsm_low      BYTE      $EF     'maximum data block number (low_byte)
:dsm_high     BYTE      $01     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $00     'total number of directory entries (high byte)
:al0          BYTE      $F0     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $06     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)
                                'N.B. SS must be 128 for CP/M 2 can be varied for CP/M 3 hard disks.

dpb_7                           '440K RAM disk as HDSK params
:spt_low      BYTE      $20     'sectors per track (low byte)
:spt_high     BYTE      $00     'sectors per track (high byte)
:bsh          BYTE      $04     'data allocation Block SHift factor
:blm          BYTE      $0F     'data allocation block mask
:exm          BYTE      $00     'EXtent Mask
:dsm_low      BYTE      $DF     'maximum data block number (low_byte)  $006F=111*32*128=454,656=444KB <======??????
:dsm_high     BYTE      $00     'maximum data block number (high_byte)
:drm_low      BYTE      $FF     'total number of directory entries (low byte)
:drm_high     BYTE      $00     'total number of directory entries (high byte)
:al0          BYTE      $F0     'determine reserved directory blocks
:al1          BYTE      $00     'determine reserved directory blocks
:cks_low      BYTE      $00     'size of directory ChecK vector (low byte)
:cks_high     BYTE      $00     'size of directory ChecK vector (high byte)
:off_low      BYTE      $00     'number of reserved tracks (offset) (low byte)
:off_high     BYTE      $00     'number of reserved tracks (offset) (high byte)
:psh          BYTE      $00     'Physical record SHift factor, CP/M 3
:phm          BYTE      $00     'PHhysical record Mask, CP/M 3
:ss_low       BYTE      $80     'Sector Size (low byte)
:ss_high      BYTE      $00     'Sector Size (high byte)
                                'N.B. SS must be 128 for CP/M 2 can be varied for CP/M 3 hard disks.


VAR     

'Parameter list for CPU objects, do not change order.
long reg_base                   'Address (in HUB)of Z80 register file
long io_base                    'Address (in prop RAM) of the I/O contol interface

word err                        'Error code return value

long switch_reg 
long simhcmd

#ifdef TriBladeProp
'For SRAM CPM Disk Loading
  long  filename
  long  srcname
  byte  buffer[16]
  long  sectorsize
'For microSD routines

  long  ioControl[2]               ' SD parameters                     
'  long  drive_base[18]             ' SD base address of drives 0..15 (A:..L:) & ram,ram,rom sector 0
  long  blockno                    ' blockno within file
  long  sdsector                   ' sector no 
  long  count                      ' count of chars to read
  byte  buff[8192]                 ' SD buffer
#endif

#ifdef FloppySupport
  byte altair_drive                'The Altair disk drive currently selected, -1 = none
  byte altair_track[16]            'The current track position on each Altair disk drive
  byte altair_sector               'The current sector on an altair disk
  byte altair_bidx                 'Byte index into Altair disk sectors
  byte altair_write_enable         'Bytes can be written to the Altair disk when = 1
  long sd_block
#endif

  byte hdsk_reset_count            'Counts number of consequiteve hdsk reset commands.
  byte hdsk_current_command        'The hdsk command currently in progress
  byte hdsk_last_command           'The last complete hdsk command issued
  byte hdsk_command_pos            'Counts through the outputed command parameter bytes (OUTs)
  byte hdsk_reply_pos              'Counts hdisk port reply bytes (INs)
  long hdsk_dpb_ptr                'Pointer into DPB tables.
  byte hdsk_drive                  'The hdsk drive to read/write with this command
  byte hdsk_sector                 'The hdsk sector to read/write with this command
  word hdsk_track                  'The hdsk track to read/write with this command
  word hdsk_dma                    'The DMA address to read/write from with this command

#ifdef FloppySupport
  long floppy_base[8]              '\\ SD base address of floppy drives 0..7 (A:...H:)
#endif
  long drive_base[18]              '// SD base address of drives 0..15 (A:..L:) + boot code
  long sector_number               'Sector of a disk starting from 0
  long sd_block_number             'SD card 512 byte block number starting from 0
  long block_offset                'Byte offset of a sector within a block
  long current_block_number        'The SD block currently in the disk buffer
  byte disk_buff[512]              'The disk I/O buffer holds one SD card block.
  long xxxx                        'requuired for overflow of disk_buff in floppy mode !!!

PUB start | i

  waitcnt(clkfreq*3 + cnt)      'delay (so you can get PropTerminal active)     <=== 3 secs

  current_block_number := -1    'No SD block read into buffer yet so set invalid value

  'Start the terminal
  UART.start(31,30,0,baud)      'start the uart Tx and Rx and set the baud rate

#ifdef TriBladeProp
  UART.str(string("ZiCog 1.0_rc_6 on the TriBladeProp v0.123"))
  crlf
#else
#warn No platform defined
#endif

  reg_base        := @register_file
  io_base         := @io_cmd_block
  switch_reg      := $08                                '$08 is the switch setting SIMH recommends for 4k Basic etc to run
'''???  firsttime       := 0                                  'if 0 display first disk sector read by Z80

#ifdef TriBladeProp

'------------------------------------------------------------------------------
' Start the disks (from SD card)
' Load SRAM with Z80 RAM (64KB) & optional CPM Ram Disk (448KB/960KB)
'------------------------------------------------------------------------------
  UART.str (string("Starting Hdisks... "))
  crlf
  blockno := 0
  count := 8192
  FindSDblock                                           'set the SD card file/drive base(s)

  UART.str(string("Loading SRAM... "))
  crlf
  LoadSRAM                                              'clear 64KB sram, diskboot at $FF00, fill >64KB with $E5

#elseifdef PropDemoBoard

'------------------------------------------------------------------------------
' Start the disks (from SD card)
'------------------------------------------------------------------------------
  UART.str (string("Starting Hdisks (A:B:C:D:E:F:G:H) ... "))
  crlf
  'set the drive base for each drive on SD (sector 0)
  repeat i from 0 to 7
    drive_base[i] := hdsk_base_block + (i * 8192 * 2)   'fixed 8 x 8MB drives for now

'  err := \sd.start(0)
  err := \sd.start(spiDO,spiClk,spiDI,spiCS)            'set SD card pins
  DisplayError(err)

#endif

#ifdef BOOT_FROM_SD_SPIN
  UART.str(string("Loading Z80 ram..."))
  LoadZ80Ram
#endif

'------------------------------------------------------------------------------
' ZiCog Emulation starts here...
'------------------------------------------------------------------------------
#ifdef CPU_Z80
  UART.str(string("Starting Z80 emulation..."))
#else
  UART.str(string("Starting 8080 emulation..."))
#endif
  crlf
  err := \cpu.start(@reg_base)
  DisplayError(not err)

  'Start the Altair I/O simulation
  service_io

'------------------------------------------------------------------------------
'
'------------------------------------------------------------------------------



'==============================================================================
' TriBladeProp Routines
'==============================================================================
#ifdef TriBladeProp

PRI FindSDblock | i, n, r                               'read blockno of file for count into buffer

  r := sd.startSD(@ioControl)                           'start the SD routines
  CheckError(not r)
  r := sd.mount_explicit(spiDO,spiClk,spiDI,spiCS)      'mount the SD
  CheckError(r)

#ifdef FloppySupport
  'floppies A:,B:,C:,D:,E:,F:,G:,H:
  FindDriveBase(0, string("A.DSK"))                     'set the SD card drive base(s)
  FindDriveBase(1, string("B.DSK"))                     'set the SD card drive base(s)
  FindDriveBase(2, string("C.DSK"))                     'set the SD card drive base(s)
  FindDriveBase(3, string("D.DSK"))                     'set the SD card drive base(s)
  FindDriveBase(4, string("E.DSK"))                     'set the SD card drive base(s)
  FindDriveBase(5, string("F.DSK"))                     'set the SD card drive base(s)
  FindDriveBase(6, string("G.DSK"))                     'set the SD card drive base(s)
  FindDriveBase(7, string("H.DSK"))                     'set the SD card drive base(s)
  longmove(@floppy_base,@drive_base,8)                  'copy the offsets to the floppy base(s)

  'hard disks I:,J:
  FindDriveBase(8, string("I.DSK"))                     'set the SD card drive base(s)   8MB H/Disk
  FindDriveBase(9, string("J.DSK"))                     'set the SD card drive base(s)   8MB H/Disk
  longmove(@drive_base[0],@drive_base[8],2)             'copy to offsets to the hard drive(s)

  'other files Q:,R:
  FindDriveBase(17,string("DSKBOOT.COM"))               'set the SD card drive base(s)    $FF00 for boot floppy

#else
  'HDisks A:,B:,C:,D:,E:,F:,G:,H:
  FindDriveBase(0, string("A.DSK"))                     'set the SD card drive base(s)  8MB CPM2.2
  FindDriveBase(1, string("B.DSK"))                     'set the SD card drive base(s)  8MB CPM2.2
  FindDriveBase(2, string("C.DSK"))                     'set the SD card drive base(s)  8MB CPM2.2
  FindDriveBase(3, string("D.DSK"))                     'set the SD card drive base(s)  8MB CPM2.2
  FindDriveBase(4, string("E.DSK"))                     'set the SD card drive base(s)  8MB CPM2.2
  FindDriveBase(5, string("F.DSK"))                     'set the SD card drive base(s)  8MB CPM2.2
  FindDriveBase(6, string("G.DSK"))                     'set the SD card drive base(s)  8MB CPM2.2
  FindDriveBase(7, string("H.DSK"))                     'set the SD card drive base(s)  8MB CPM2.2

  FindDriveBase(17,string("ZIBOOT.COM"))                'set the SD card drive base(s)    $FF00 for boot hdisk

#endif

  r := sd.stopSDCard                                    'stop SD card, but do not stop cog
  CheckError(r)


PRI FindDriveBase(n, st) | r, fsize                     'set the SD drive_base for the drive no
  UART.tx($41 + n)
  UART.tx(":")                                          'A:...
  UART.str(st)

  r := sd.popen(st,"r")                                 'Open file for read (blank image)
  CheckError(r)
#ifdef fsrw_new
  drive_base[n] := sd.getSDdatablock                    'get the sectorno of the file (returns the sector+1)
  fsize := sd.get_filesize
'  if fsize < 512
'    drive_base[n] := drive_base[n] - 1                  'correct for < 1 sector
#else
  r := sd.pread(@buff,1)                                'get a char to find the sector address ****
  CheckError(r)
  drive_base[n] := ioControl[0] - 1                     'get the sectorno of the file (returns the sector+1)
#endif

#ifdef RamHDisk
  if n == 7                                             'H: is the RamDisk
    UART.str(string(" <-RamDisk"))
#endif
  if n == 17
    UART.str(string(" <-bootcode"))

  uart.tx(" ")
  uart.dec(drive_base[n])
  uart.tx(" ")
  uart.dec(fsize)

  crlf


PRI LoadSRAM | i, j, r, b1, b2

  count := 8192                                         'use 8KB blocks
  b1 := cpmsize / count                                 '8KB blocks in 64KB CPM RAM
  b2 := sramsize / count                                '8KB blocks in 512KB/1MB SRAM

'clear the whole 64KB sram first (need a command in tbp2 to do this)
  r := tbp2.start                                       'start the triblade2 driver
  CheckError(r)

  repeat i from 0 to count - 1                          'set 8KB block = $00
    buff[i] := 0

  repeat i from 0 to b1 -1                              '0 to 7: write 8 x 8KB blocks =$00 to sram
    r := tbp2.DoCmd("W", @buff, i * count, count)       'write to sram
    CheckError(r)

  buff[0] := $C3                                        'force jmp to $FF00 at $0000 (faster)
  buff[1] := $00
  buff[2] := $FF
  r := tbp2.DoCmd("W", @buff, 0, 3)                     'write to sram
  CheckError(r)

  r := tbp2.idle                                        'give up the bus
  CheckError(r)

  r := sd.initSDCard(spiDO,spiClk,spiDI,spiCS)          'init SD card, but do not restart cog
  CheckError(r)

#ifdef fsrw_new
  uart.tx("%")
#endif

'read the DSKBOOT_.ROM file (256 bytes)
'read the ZIBOOT  .COM file (256 bytes)
'  r := sd.readSDCard(drive_base[16], @buff, 256)        'get 256 bytes   "DSKBOOT_.ROM"
  r := sd.readSDCard(drive_base[17], @buff, 256)        'get 256 bytes   "ZIBOOT  .COM"
  CheckError(r)

#ifdef fsrw_new
  uart.tx("&")
#endif

  r := sd.stopSDCard                                    'stop SD card, but do not stop cog
  CheckError(r)

#ifdef fsrw_new
'DISPLAY THE DATA READ
  uart.str(string("data"))
  crlf
  repeat j from 0 to 15
    repeat i from 0 to 15
      UART.hex(buff[(j*16)+i],2)
      UART.tx(" ")
    crlf
  crlf

  repeat '<=========================
#endif

'copy the 256 bytes to $FF00-$FFFF in sram
  r := tbp2.active                                      'activate the bus
  CheckError(r)

  r := tbp2.DoCmd("W", @buff, $FF00, 256)               'write to sram
  CheckError(r)

'now set the buffer to "$E5" and fill the remainder of sram for use as empty RamDisk
  repeat i from 0 to count - 1                          'set 8KB block = $E5
    buff[i] := $E5

  repeat i from b1 to b2 - 1                            '8 to 63/127: write 56/120 x 8KB blocks of sram above first 64KB
    r := tbp2.DoCmd("W", @buff, i * count, count)       'write to sram
    CheckError(r)

' ^ leave the tbp2 bus active (U23 sram selected and latched)

PRI show(s) | i, addrx
  addrx := blockno * count
  UART.hex(addrx,6)
  UART.tx(" ")
  repeat i from 0 to 15                                 'only display first xx chars per block
#ifdef Show_Hex
    UART.hex(buff[i],2)
    UART.tx(" ")
#else
    UART.tx(buff[i])
#endif
  crlf


PRI CheckError(r)
  if r < 0
    UART.str(string("***Error*** $"))
    UART.hex(r,2)
    UART.str(string(" ***HALTED***"))
    crlf
    sd.stop
    tbp2.stop
    repeat                                              '<==============  LOOP HERE

'------------------------------------------------------------------------------
#endif
'==============================================================================

PRI DisplayError(errx)
  if errx
    UART.str(string(" ***Failed***"))
  else
    UART.str(string("Passed, please wait..."))
  crlf


#ifdef BOOT_FROM_SD_SPIN
PRI LoadZ80Ram | bootBlock, sect, z80_address
' Reads 88 sectors (22 SD blocks) from SD to Z80 RAM.
' FIXME: Should allow selectable boot disks
  block_offset := 0
  z80_address := 0
  repeat bootBlock from 2 to 23
#ifdef PropDemoBoard
    err := \sd.readblock(bootBlock, @disk_buff)
#elseifdef TriBladeProp
    'Cluso something here perhaps
    err := -5  'force error!!
#endif
    if err
      UART.str(string("Error hdisk boot read"))
      return
    repeat sect from 0 to 3
      block_offset := sect * 128
      cpu.dma_write (z80_address, (@disk_buff) + block_offset)
      z80_address += 128
#endif


'------------------------------------------------------------------------------
' Service I/O
'------------------------------------------------------------------------------
PRI service_io
'Altair I/O simulation service loop
  repeat
    case io_command
      io_cmd_break:                                     'CPU break
        on_break    
      io_cmd_in:                                        'An input operation
        on_input
      io_cmd_out:                                       'An output operation
        on_output

PRI on_break
'Handle CPU break condition
'No, this never existed in an Altair (or SIMH simulator)
  print_regs
  'waitcnt (cnt + 80_000)
  repeat while UART.rx <> $20                           'wait for 'space'
  io_command := 0


PRI on_input
'Handle Z80 IO port reads
  case io_port
    srport:                                             'Altair front panel switches register
      in_srport
    simhport:                                           'SIMH control port
      in_simhport          
    constat:                                            'Console status port.
      in_constat
    condata:                                            'Console data port
      in_condata
#ifdef FloppySupport
    statin:                                             'Altair disk status port
      in_statin
    secpos:                                             'Altair disk sector position port
      in_secpos
    dskread:                                            'Altair disk read port
      in_dskread
#endif
    n8vem_uart_test:                                    ' test port $6D
      in_test_n8vem                                     ' is a byte ready returns 0 for no byte, 1 for a byte
    n8vem_uart_in:                                      ' ask for input from port $68
      in_uart_n8vem                                     ' get the byte
    hdskport:
      in_hdskport                                       'SIMH style hard disk port
    other:
      'UART.str(string("Read from unimplemented I/O Port &H"))
      'UART.hex(io_port,2)
      'crlf
      io_data := 0
  io_command := 0                                       'Set flag ready for next io

PRI in_test_n8vem
'  for xmodem tests if any input from the serial port
' input present = bit 0. Output ready to go = bit 5 (always high). bit1=0 (carrier detect)
  if UART.rxavail
    io_data :=%00100001                                 'yes
  else
    io_data :=%00100000                                 'no

PRI in_uart_n8vem
  if UART.rxavail 
    io_data := UART.rx                                  ' get the byte
  else
    io_data := "?"                                      '<=== at least show a character to indicate a problem

PRI in_srport
'Handle reads from the Altair front panel switch port
  io_data := switch_reg
  UART.str(string("SR port read"))
  crlf

PRI in_simhport
'Handle reads from the SIMH simulator status port
  if simhcmd == hasbankedmemory
    io_data :=  1
    simhcmd := 0

PRI in_constat
'Handle reads from the console status port
  if UART.rxavail                                       'Is there a character in the buffer
    io_data := %00000011                                'Return flag to say a character is available
  else
    io_data := %00000010                                'No: Return no char ready for input

PRI in_condata
'Handle reads from the console data port (N.B. No wait)
  io_data := UART.rxcheck                               'Yes: Return a character (or junk if none available)


#ifdef FloppySupport
PRI in_statin
'Handle reads from the disk status port
  io_data := 0
  'if altair_write_enable == 1                          'Ready to accept data ?
    'data |= %00000001
    'data |= %00000010                                  'Head movement is always allowed !
    'data |= %00000100                                  'Head is loaded for read/write
  if altair_track[altair_drive] <> 0                    'Now on track zero ?
    io_data |= %01000000
    'data |= %10000000                           'New byte ready to read


PRI in_secpos
'Handle reads from the floppy disk sector position port
  altair_sector := altair_sector + 1
  if altair_sector > 31
    altair_sector := 0
  io_data := (altair_sector << 1) '| 1                  'Return sector and positioned bit
                                                        'Note: MPR BIOS expects positioned bit = 0 when positioned
                                                        'despite what the BIOS comments say about it.
  altair_bidx := 0

PRI in_dskread | bix, r, iy
'Handle reads from the floppy disk read data port
  if altair_bidx == 0
    'Read a new block from SD card
'   '3pre & 6post bytes for 137 byte sector not checked, so skip them RR2000819 and read into @disk_buff[3]
'----------------------------------------------------------------------------------------------
    r := tbp2.idle                                      'give up the bus
    CheckError(r)

    r := sd.initSDcard(spiDO,spiClk,spiDI,spiCS)        'init the SD
    CheckError(r)

    ' this method relies on the SD file using contiguous sectors !!! Only the first 128 bytes of 512 bytes used
    sd_block := (floppy_base[altair_drive]) + (altair_track[altair_drive] * 32) + altair_sector
    err := sd.readSDCard(sd_block, @disk_buff[3], 128)  'read a sector
'   CheckError(err)

    r := sd.stopSDcard                                  'stop the SD (free up pins, but keep cog running)
    CheckError(r)

    r := tbp2.active                                    'activate the bus
    CheckError(r)
'----------------------------------------------------------------------------------------------
    if err
      UART.str(string("Disk Read "))
      DisplayError(err)
      DisplayDiskDTS

  io_data := disk_buff[altair_bidx]

  if altair_bidx < asecsiz
    altair_bidx++

#endif


PRI hdsk_check_dts : status
'Check that selected hdsk drive, sector and track are within limits
  if (hdsk_drive < nhdisks) and (hdsk_track =< 2048) and (hdsk_sector =< 128)
    status := $00
  else
    UART.str(string("Ilegal hdsk access"))
    DisplayHDiskDTS
    status := $FF


PRI refresh_hd_cache | r
'Ensure that the sector number in use is in the current SD block in memory.
'If not load that block to disk buff from SD card.

#ifdef PropDemoBoard

{
  'WARNING This is jumping over 8MB chunks of SD card.
  sector_number := hdsk_drive * (32 * 2048)           'Skip on to the first sector of the selected drive
  sector_number += (hdsk_track * 32)                  'Find correct block for the selected track..
  sector_number += hdsk_sector                        '..and sector
  sd_block_number := sector_number >> 2               'Determine the SD block of this sector..
  sd_block_number += hdsk_base_block                  '..starting at first block of first HD is SD.
  block_offset := (sector_number & %11) * 128         'Determine the position of this sector within the block
}

  sector_number   := (hdsk_track * 32) + hdsk_sector                 '128 byte sector no
  sd_block_number := drive_base[hdsk_drive] + (sector_number >> 2)   'base + sector/4 (because 4x128 per 512 block)
  block_offset    := (hdsk_sector & %11) * 128                       'position of this sector within the block


  if sd_block_number <> current_block_number                         'If the block is not in disk buffer then load it
'    err := \sd.readblock(sd_block_number, @disk_buff + block_offset) 'Read the 512 byte block !
    err := \sd.readblock(sd_block_number, @disk_buff)                'Read the 512 byte block !
'error ???? overflows @disk_buff                      ^^^^^^^^^^^^^^    <=================================

'    uart.dec(sd_block_number)
'    crlf

#elseifdef TriBladeProp
{  <++++++++++++++
'----------------------------------------------------------------------------------------------
#ifdef RamHDisk
'----------------------------------------------------------------------------------------------
  if hdsk_drive == 1        not debugged _rr114                            'J: is RamDrive

    sd_block_number := (hdsk_track * 32) + hdsk_sector
    err := tbp2.DoCmd("R", @disk_buff, sd_block_number*128 + cpmsize, 128)

  else

#else
'----------------------------------------------------------------------------------------------
#endif
} '<++++++++++

  sector_number := (hdsk_track * 32) + hdsk_sector    'Find correct block for the selected track and sector                        '..and sector
  sd_block_number := sector_number >> 2               'Determine the SD block of this sector..
  sd_block_number += drive_base[hdsk_drive]           'get the first sector of the selected drive
  block_offset := (sector_number & %11) * 128         'Determine the position of this sector within the block

  if sd_block_number <> current_block_number          'If the block is not in disk buffer then load it

    r := tbp2.idle                                      'give up the bus
    CheckError(r)

    r := sd.initSDcard(spiDO,spiClk,spiDI,spiCS)        'init the SD
    CheckError(r)

    ' this method relies on the SD file using contiguous sectors !!! Packed 4 x 128 bytes per 512 byte sectors
    err := \sd.readSDCard(sd_block_number, @disk_buff, 512)     'read the block! (complete 512 bytes)
'    err := \sd.readSDCard(sd_block_number, @disk_buff + block_offset, 512)     'read the block! (complete 512 bytes)
'error ???? overflows @disk_buff                      ^^^^^^^^^^^^^^    <=================================
'   CheckError(err)

    r := sd.stopSDcard                                  'stop the SD (free up pins, but keep cog running)
    CheckError(r)

    r := tbp2.active                                    'activate the bus
    CheckError(r)

#endif

    if err
      UART.str(string("Error hdisk cache read"))
      DisplayError(err)
      DisplayHDiskDTS
      io_data := $FF                                  'Return error code
    else
      current_block_number := sd_block_number         'Make this new block current


PRI in_hdskport_read | r
'Handle read from hdskport during a disk read command
'This read triggers the actual disk sector transfer and must return an error code (To the Z80 not the calling Spin)
'N.B. sectors are mapped to blocks 4 to 1

  if hdsk_check_dts                                   'Check drive, track and sector are sane
    io_data := $FF                                    'Return error if not
    return

  refresh_hd_cache

  'Virtual "DMA" the sector into DMA buffer address in Z80 RAM

#ifdef PropDemoBoard

  cpu.dma_write(hdsk_dma, @disk_buff + block_offset)

#elseifdef TriBladeProp

'#ifdef RamHDisk ?????     <=============================  code to go here for RamDisk

  r := tbp2.DoCmd("W", @disk_buff + block_offset, hdsk_dma, 128) 'Virtual "DMA" the sector into DMA buffer in SRAM
  CheckError(r)

#endif

  io_data := $00                                        'Return OK.
  hdsk_last_command := hdskNull                         'Command completed


PRI in_hdskport_write | r
'Handle read from hdskport during a disk write command
'This read triggers the actual disk sector transfer and must return an error code(To the Z80 not the calling Spin)
'N.B. sectors are mapped to blocks 4 to 1

  if hdsk_check_dts                                     'Check drive, track and sector are sane
    io_data := $FF                                      'Return error if not
    return

  refresh_hd_cache

  'Virtual "DMA" the sector from DMA buffer address in Z80 RAM
#ifdef PropDemoBoard

  cpu.dma_read (hdsk_dma, @disk_buff + block_offset)
  err := \sd.writeblock(sd_block_number, @disk_buff)    'Write the block! (complete 512 bytes)

#elseifdef TriBladeProp

'#ifdef RamHDisk ?????     <=============================  code to go here for RamDisk

  'read the block in from Z80 sram
  r := tbp2.DoCmd("R", @disk_buff + block_offset, hdsk_dma, 128) 'Virtual "DMA" the sector from DMA buffer from SRAM
  CheckError(r)

  r := tbp2.idle                                        'give up the bus
  CheckError(r)

  r := sd.initSDcard(spiDO,spiClk,spiDI,spiCS)          'init the SD
  CheckError(r)

  ' this method relies on the SD file using contiguous sectors !!! Packed 4 x 128 bytes per 512 byte sectors
  err := \sd.writeSDCard(sd_block_number, @disk_buff, 512)   'write the block! (complete 512 bytes)
' CheckError(err)

  r := sd.stopSDcard                                    'stop the SD (free up pins, but keep cog running)
  CheckError(r)

  r := tbp2.active                                      'activate the bus
  CheckError(r)
'----------------------------------------------------------------------------------------------

#endif

  if err
    UART.str(string("Error hdisk write "))
    DisplayError(err)
    DisplayHDiskDTS
    io_data := $FF                                      'Return error code
  else
    io_data := $00                                      'Return OK.

  hdsk_last_command := hdskNull                         'Command completed


PRI in_hdskport_param
'Handle read from hdskport during disk parameters command
'Simply return the next param byte from a DPB.
  'UART.str(string("hdskparam "))
  'UART.dec(hdsk_reply_pos)
  'UART.str(string(" = "))
  io_data := hdsk_dpb[hdsk_reply_pos++]
  io_data := BYTE[hdsk_dpb_ptr++]
  hdsk_reply_pos++
  'UART.hex(io_data,2)
  'crlf
  if hdsk_reply_pos == 19
    hdsk_reply_pos := 0
    hdsk_last_command := hdskNull                   'Command completed
    UART.str(string("command completed "))
    crlf


PRI in_hdskport
'Handle reads from the hard disk interface port
'Handling is dispatched to an appropriate function depending on the disk operation currently in progress
  case hdsk_last_command
    hdskRead:
      in_hdskport_read
    hdskWrite:
      in_hdskport_write
    hdskParam:                                          'Current command is get disk params so just return successive parameter bytes from DPB.
      in_hdskport_param
    other:
      UART.str(string("HDSK ERROR 5"))
      crlf


PRI on_output
'Handle Z80 IO port writes
  case io_port
    ledport:                                            'An 8 bit parallel output with LEDs attached
      out_ledport
    simhport:                                           'SIMH control port
      out_simhport
    condata:                                            'Console data port
      out_condata
    constat:                                            'Console status port
      out_constat
    punstat:
      out_punstat                                       'Punch control port
#ifdef FloppySupport
    selout:                                             'Altair disk controller select port
      out_selout
    dskcon:                                             'Altair disk function port
      out_dskcon
    dskwrit:                                            'Altair disk write port
      out_dskwrit
#endif
    n8vem_uart_out:                                     '&H68
      out_n8vem_byte
    hdskport:
      out_hdskport                                      'SIMH style hard disk port
    other:
      'UART.str(string("Write to unimplemented I/O Port &H"))
      'UART.hex(io_port,2)                               'print the port number for debugging
      'crlf
  io_command := 0                                       'Set flag ready for next io

PRI out_ledport
'Handle writes to the LED port
'No, SIMH never had this. I added it for the Propeller demo board.
  'outa[16..23] := io_data                              'Output the port data  (if re-enabled do #ifdef)
  UART.str(string("LED port = "))
  UART.hex(io_data,2)
  crlf


PRI out_simhport
'Handle writes to the SIMH simulator control port


PRI out_condata
'Handle writes to the console data output port
    UART.tx(io_data)


PRI out_constat
'Handle writes to the console command port
  if io_data == creset
    'UART.str(string("Console reset"))
    'crlf
  else
    UART.str(string("Unknown console status write $"))
    UART.hex(io_data,2)
    crlf


PRI out_punstat
'Handle writes to the punch device command port
  if io_data == creset
    'UART.str(string("Punch reset"))
    'crlf
  else
    UART.str(string("Unknown punch status write $"))
    UART.hex(io_data,2)
    crlf             


#ifdef FloppySupport
PRI out_selout
'Handle writes to the floppy disk drive select port
  if io_data & %10000000
    altair_drive := -1
  else
    altair_drive := io_data & %00001111

PRI out_dskcon
'Handle writes to the floppy disk control port
  if (io_data & cstepin) > 0                            'Step head in command ?
    if altair_track[altair_drive] < tracks
      altair_track[altair_drive]++
  if (io_data & cstepot) > 0                            'Step head out command ?
    if altair_track[altair_drive] > 0
      altair_track[altair_drive]--

  if (io_data & cload) > 0                              'Load head to drive surface command ?

  if (io_data & cuload) > 0                             'Unload head from drive surface command ?

  if (io_data & cwrseq) > 0                             ' 'start write enable sequence' command
    altair_bidx := 0
    altair_write_enable := 1


PRI out_dskwrit | r
'Handle writes to the floppy disk write data port
  if altair_bidx < asecsiz
    disk_buff[altair_bidx++] := io_data
    if altair_bidx == asecsiz
      'Write the buffer to the SD card or RamDisk
'   '3pre & 6post bytes for 137 byte sector not checked, so skip them RR2000819 and write from @disk_buff[3]
'----------------------------------------------------------------------------------------------
      r := tbp2.idle                                    'give up the bus
      CheckError(r)

      r := sd.initSDcard(spiDO,spiClk,spiDI,spiCS)      'init the SD
      CheckError(r)

      ' this method relies on the SD file using contiguous sectors !!! Only the first 128 bytes of 512 bytes used
      sd_block := (floppy_base[altair_drive]) + (altair_track[altair_drive] * 32) + altair_sector
      err := sd.writeSDCard(sd_block, @disk_buff[3], 128) 'write a sector
'     CheckError(err)

      r := sd.stopSDcard                                'stop the SD (free up pins, but keep cog running)
      CheckError(r)

      r := tbp2.active                                  'activate the bus
      CheckError(r)
'----------------------------------------------------------------------------------------------
      if err
        UART.str(string("Disk Write "))
        DisplayError(err)
        DisplayDiskDTS

      altair_write_enable := 0

#endif


PRI out_n8vem_byte
  UART.tx(io_data)   ' send out the data to the port (for xmodem)


PRI hdsk_get_rw_params
  case hdsk_command_pos
    1:
      hdsk_drive := io_data                             '1st byte is the drive number
      hdsk_command_pos++
    2:
      hdsk_sector := io_data                            '2nd byte is the sector number
      hdsk_command_pos++
    3:
      hdsk_track := io_data                             '3rd byte is LOW of the track number
      hdsk_command_pos++
    4:
      hdsk_track := hdsk_track + (io_data << 8)         '4th byte is HIGH of the track number
      hdsk_command_pos++
    5:
      hdsk_dma := io_data                               '5th byte is LOW of the DMA address
      hdsk_command_pos++
    6:
      hdsk_dma := hdsk_dma + (io_data << 8)             '6th (and last) byte is HIGH of the DMA address
      hdsk_command_pos++
      hdsk_last_command := hdsk_current_command         'OUT's for current command completed
      hdsk_current_command := hdskNull
    other:
      UART.str(string("ERROR HDSK 1"))
      crlf
      repeat


PRI out_hdskport_param
'Handle writes to the hdskport when get params command is in progress
  case hdsk_command_pos
    1:
      hdsk_drive := io_data
      'UART.str(string("Hdsk drive = "))
      'UART.dec(hdsk_drive)
      'crlf
      hdsk_command_pos++
      hdsk_reply_pos := 0
      hdsk_dpb_ptr := @hdsk_dpb + (hdsk_drive * 19)     'Point to correct DPB for this drive
      hdsk_last_command := hdsk_current_command         'OUT's for current command completed
      hdsk_current_command := hdskNull
    other:
      UART.str(string("HDSK ERROR 4"))
      crlf
      repeat


PRI out_hdskport_null
'Handle writes to the hdskport when no command is in progress
'io_data should be a new command byte.
  case io_data
    hdskRead:
      'UART.str(string("hdsk read"))
      'crlf
      hdsk_current_command := hdskRead
      hdsk_command_pos := 1
    hdskWrite:
      'UART.str(string("hdsk write"))
      'crlf
#ifdef ShowDiskWrites
      UART.tx(".")                                      'indicates writing as takes long time (looks like hang)
#endif
      hdsk_current_command := hdskWrite
      hdsk_command_pos := 1
    hdskParam:
      'UART.str(string("hdsk param"))
      'crlf
      hdsk_current_command := hdskParam
      hdsk_command_pos := 1
    hdskReset:
      'UART.str(string("hdsk reset"))
      'crlf
      hdsk_current_command := hdskNull
      hdsk_last_command := hdskNull
      hdsk_command_pos := 1
      hdsk_reply_pos := 0
    other:
      UART.str(string("HDSK ERROR 2"))
      crlf
      repeat


PRI out_hdskport
'Handle writes to the SIMH style hard disk port
'Handling is dispatched to an apropriate function depending
'on the command currently in progress.
'FIXME: RESET should be checked on every write no matter the command in progress.
  case hdsk_current_command
    hdskRead:
      hdsk_get_rw_params
    hdskWrite:
      hdsk_get_rw_params
    hdskParam:
      out_hdskport_param
    hdskReset:
      UART.str(string("HDSK This cannot happen!"))
      crlf
      repeat
    hdskNull:
      out_hdskport_null
    other:
      UART.str(string("HDSK ERROR 3"))
      crlf
      repeat


PRI DisplayHDiskDTS
  UART.str(string("D = "))
  UART.hex(hdsk_drive, 2)
  UART.str(string(" T = "))
  UART.hex(hdsk_track, 4)
  UART.str(string(" S = "))
  UART.hex(hdsk_sector, 2)
  UART.str(string(" DMA = "))
  UART.hex(hdsk_dma, 4)
  crlf

#ifdef FloppySupport
PRI DisplayDiskDTS
  UART.str(string("D = "))
  UART.hex(altair_drive, 2)
  UART.str(string(" T = "))
  UART.hex(altair_track, 4)
  UART.str(string(" S = "))
  UART.hex(altair_sector, 2)
  crlf
#endif

PRI print_regs
'Print the Z80 registers
  UART.str(string("PC="))
  UART.hex(pc_reg, 4)
  UART.tx(" ")
'  UART.hex(byte[ram_base + pc_reg], 2)

  UART.str(string(" AF="))
  UART.hex(af_reg, 4)
      
  UART.str(string(" BC="))
  UART.hex(bc_reg, 4)
      
  UART.str(string(" DE="))
  UART.hex(de_reg, 4)

  UART.str(string(" HL="))
  UART.hex(hl_reg, 4)
      
  UART.str(string(" IX="))
  UART.hex(ix_reg, 4)

  UART.str(string(" IY="))
  UART.hex(iy_reg, 4)

  UART.str(string(" SP="))
  UART.hex(sp_reg, 4)
  crlf


PRI crlf
  UART.tx(13)
  UART.tx(10)
